@@ -99,6 +99,13 @@ GLogLevelFlags  g_log_set_fatal_mask    (const gchar    *log_domain,
                                          GLogLevelFlags  fatal_mask);
 GLogLevelFlags  g_log_set_always_fatal  (GLogLevelFlags  fatal_mask);
 
+/* internal */
+void	_g_log_fallback_handler	(const gchar   *log_domain,
+				 GLogLevelFlags log_level,
+				 const gchar   *message,
+				 gpointer       unused_data);
+
+
 #ifndef G_LOG_DOMAIN
 #define G_LOG_DOMAIN    ((gchar*) 0)
 #endif  /* G_LOG_DOMAIN */
@@ -193,7 +200,7 @@ GPrintFunc      g_set_printerr_handler  (GPrintFunc      func);
 #ifdef __GNUC__
 
 #define g_assert(expr)			G_STMT_START{		\
-     if (!(expr))						\
+     if (expr) { } else						\
        g_log (G_LOG_DOMAIN,					\
 	      G_LOG_LEVEL_ERROR,				\
 	      "file %s: line %d (%s): assertion failed: (%s)",	\
@@ -213,7 +220,7 @@ GPrintFunc      g_set_printerr_handler  (GPrintFunc      func);
 #else /* !__GNUC__ */
 
 #define g_assert(expr)			G_STMT_START{		\
-     if (!(expr))						\
+     if (expr) { } else						\
        g_log (G_LOG_DOMAIN,					\
 	      G_LOG_LEVEL_ERROR,				\
 	      "file %s: line %d: assertion failed: (%s)",	\
@@ -245,7 +252,7 @@ GPrintFunc      g_set_printerr_handler  (GPrintFunc      func);
 #ifdef __GNUC__
 
 #define g_return_if_fail(expr)		G_STMT_START{			\
-     if (!(expr))							\
+     if (expr) { } else							\
        {								\
 	 g_log (G_LOG_DOMAIN,						\
 		G_LOG_LEVEL_CRITICAL,					\
@@ -258,7 +265,7 @@ GPrintFunc      g_set_printerr_handler  (GPrintFunc      func);
        };				}G_STMT_END
 
 #define g_return_val_if_fail(expr,val)	G_STMT_START{			\
-     if (!(expr))							\
+     if (expr) { } else							\
        {								\
 	 g_log (G_LOG_DOMAIN,						\
 		G_LOG_LEVEL_CRITICAL,					\
@@ -291,7 +298,7 @@ GPrintFunc      g_set_printerr_handler  (GPrintFunc      func);
 #else /* !__GNUC__ */
 
 #define g_return_if_fail(expr)		G_STMT_START{		\
-     if (!(expr))						\
+     if (expr) { } else						\
        {							\
 	 g_log (G_LOG_DOMAIN,					\
 		G_LOG_LEVEL_CRITICAL,				\
@@ -303,7 +310,7 @@ GPrintFunc      g_set_printerr_handler  (GPrintFunc      func);
        };				}G_STMT_END
 
 #define g_return_val_if_fail(expr, val)	G_STMT_START{		\
-     if (!(expr))						\
+     if (expr) { } else						\
        {							\
 	 g_log (G_LOG_DOMAIN,					\
 		G_LOG_LEVEL_CRITICAL,				\
